from __future__ import print_function

import pytest
from mockito import spy2, unstub, verify, verifyZeroInteractions

from . import TealConfig, TealLevel, TealPrint


@pytest.mark.parametrize(
    "name,level,function_tuple",
    [
        (
            "Prints nothing when level is none",
            TealLevel.none,
            [
                (TealPrint.error, False),
                (TealPrint.warning, False),
                (TealPrint.info, False),
                (TealPrint.verbose, False),
                (TealPrint.debug, False),
            ],
        ),
        (
            "Prints error when level is error",
            TealLevel.error,
            [
                (TealPrint.error, True),
                (TealPrint.warning, False),
                (TealPrint.info, False),
                (TealPrint.verbose, False),
                (TealPrint.debug, False),
            ],
        ),
        (
            "Prints warning+error when level is warning",
            TealLevel.warning,
            [
                (TealPrint.error, True),
                (TealPrint.warning, True),
                (TealPrint.info, False),
                (TealPrint.verbose, False),
                (TealPrint.debug, False),
            ],
        ),
        (
            "Prints info+warning+error when level is info",
            TealLevel.info,
            [
                (TealPrint.error, True),
                (TealPrint.warning, True),
                (TealPrint.info, True),
                (TealPrint.verbose, False),
                (TealPrint.debug, False),
            ],
        ),
        (
            "Prints verbose+info+warning.error when level is verbose",
            TealLevel.verbose,
            [
                (TealPrint.error, True),
                (TealPrint.warning, True),
                (TealPrint.info, True),
                (TealPrint.verbose, True),
                (TealPrint.debug, False),
            ],
        ),
        (
            "Prints everything when level is debug",
            TealLevel.debug,
            [
                (TealPrint.error, True),
                (TealPrint.warning, True),
                (TealPrint.info, True),
                (TealPrint.verbose, True),
                (TealPrint.debug, True),
            ],
        ),
    ],
)
def test_print_level(name: str, level: TealLevel, function_tuple):
    print(name)

    TealConfig.level = level

    for function, expected in function_tuple:
        spy2(TealPrint._buffer._add_to_buffer)

        function("message")

        if expected:
            verify(TealPrint._buffer, atleast=1)._add_to_buffer(...)
        else:
            verifyZeroInteractions()

        unstub()



# Exported classes
from .tealconfig import TealConfig  # lgtm[py/unused_import]
from .teallevel import TealLevel  # lgtm[py/unused_import]
from .tealprint import TealPrint  # lgtm[py/unused_import]
from .tealprintbuffer import TealPrintBuffer  # lgtm[py/unused_import]



import pytest

from . import TealConfig, TealLevel, TealPrintBuffer

indent = "".ljust(TealConfig.indent_by, TealConfig.indent_char)


class T:
    logger = TealPrintBuffer()

    @staticmethod
    def reset():
        TealConfig.level = TealLevel.info
        T.logger = TealPrintBuffer()


@pytest.mark.parametrize(
    "name,expected,function",
    [
        (
            "Not indented by default",
            "test\n",
            lambda: T.logger.info("test"),
        ),
        (
            "Indented when when pushed with same lever",
            f"{indent}test\n",
            lambda: (
                T.logger.push_indent(TealLevel.info),
                T.logger.info("test"),
            ),
        ),
        (
            "Indented when when pushed with higher lever",
            f"{indent}test\n",
            lambda: (
                T.logger.push_indent(TealLevel.warning),
                T.logger.info("test"),
            ),
        ),
        (
            "Not Indented when when pushed with same lever",
            f"test\n",
            lambda: (
                T.logger.push_indent(TealLevel.verbose),
                T.logger.info("test"),
            ),
        ),
        (
            "Indented when when pushed in info()",
            f"first\n{indent}test\n",
            lambda: (
                T.logger.info("first", push_indent=True),
                T.logger.info("test"),
            ),
        ),
        (
            "Don't de-indent after pop if level is lower than current",
            f"{indent}test\n{indent}test\ntest\n",
            lambda: (
                T.logger.push_indent(TealLevel.info),
                T.logger.push_indent(TealLevel.verbose),
                T.logger.info("test"),
                T.logger.pop_indent(),
                T.logger.info("test"),
                T.logger.pop_indent(),
                T.logger.info("test"),
            ),
        ),
        (
            "No indent after cleaning",
            f"test\n",
            lambda: (
                T.logger.push_indent(TealLevel.info),
                T.logger.clear_indent(),
                T.logger.info("test"),
            ),
        ),
        (
            "Always push indent when logging",
            "Works!\n",
            lambda: (
                T.logger.info("Works!", push_indent=True),
                T.logger.debug("Not shown", push_indent=True),
                T.logger.pop_indent(),
                T.logger.pop_indent(),
            ),
        ),
    ],
)
def test_indentation(name, expected: str, function) -> None:
    print(name)

    function()

    T.logger.buffer.seek(0)
    result = T.logger.buffer.read()
    assert expected == result

    T.reset()



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

project_slug = "tealprint"
module_name = project_slug.replace("-", "_")

setup(
    name=project_slug,
    use_scm_version=True,
    url=f"https://github.com/Senth/{project_slug}",
    license="MIT",
    author="Matteus Magnusson",
    author_email="senth.wallace@gmail.com",
    description="Prints messages to the console",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=["colored"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires=">=3.8",
)



from .teallevel import TealLevel
from .tealprintbuffer import TealPrintBuffer


class TealPrint:
    _buffer = TealPrintBuffer()

    @staticmethod
    def error(
        message: str,
        push_indent: bool = False,
        exit: bool = False,
        print_exception: bool = False,
        print_report_this: bool = False,
    ):
        """Recommendation: Only use this when an error occurs and you have to exit the program
           Prints an error message in red, can quit and print an exception.
           Also prints a "Please report this and paste the above message"

        Args:
            message (str): The message to print
            push_indent (bool): If messages after this should be indented by one level. Call pop_indent() to unindent
            exit (bool): If the program should exit after printing the error
            print_exception (bool): Set to true to print an exception
            print_report_this (bool): Set to true to add an "Please report this..." message at the end
        """
        TealPrint._buffer.error(message, push_indent, exit, print_exception, print_report_this)
        TealPrint._buffer.flush()

    @staticmethod
    def warning(message: str, push_indent: bool = False, exit: bool = False):
        """Prints an orange warning message message

        Args:
            message (str): The message to print
            push_indent (bool): If messages after this should be indented by one level. Call pop_indent() to unindent
            exit (bool): If the program should exit after printing the warning
        """
        TealPrint._buffer.warning(message, push_indent, exit)
        TealPrint._buffer.flush()

    @staticmethod
    def info(message: str, push_indent: bool = False, color: str = ""):
        """Print a message if TealPrint.level has been set to debug/verbose/info

        Args:
            message (str): The message to print
            push_indent (bool): If messages after this should be indented by one level. Call pop_indent() to unindent
            color (str): Optional color of the message
        """
        TealPrint._buffer.info(message, push_indent, color)
        TealPrint._buffer.flush()

    @staticmethod
    def verbose(message: str, push_indent: bool = False, color: str = ""):
        """Prints a message if TealPrint.level has been set to debug/verbose

        Args:
            message (str): The message to print
            push_indent (bool): If messages after this should be indented by one level. Call pop_indent() to unindent
            color (str): Optional color of the message
        """
        TealPrint._buffer.verbose(message, push_indent, color)
        TealPrint._buffer.flush()

    @staticmethod
    def debug(message: str, push_indent: bool = False, color: str = ""):
        """Prints a message if the TealPrint.level has been set to debug

        Args:
            message (str): The message to print
            push_indent (bool): If messages after this should be indented by one level. Call pop_indent() to unindent
            color (str): Optional color of the message
        """
        TealPrint._buffer.debug(message, push_indent, color)
        TealPrint._buffer.flush()

    @staticmethod
    def push_indent(level: TealLevel) -> None:
        """Add indentation for all messages, but only if level is would be shown"""
        TealPrint._buffer.push_indent(level)

    @staticmethod
    def pop_indent() -> None:
        """
        Remove indentation for all messages. Note that nothing will happen if push_indent() was
        called with a level that wouldn't be shown.

        For example:
            TealConfig.level = TealLevel.info
            TealPrint.push_indent(TealLevel.warning)
            TealPrint.info("Will be indented by 1 level")
            TealPrint.push_indent(TealLevel.verbose)
            TealPrint.info("Will still be indented by 1 level")
            TealPrint.pop_indent()
            TealPrint.info("Will still be indented by 1 level...")
            TealPrint.pop_indent()
            TealPrint.info("Not indentend")
        """
        TealPrint._buffer.pop_indent()

    @staticmethod
    def clear_indent() -> None:
        """
        Removes all indentation. Note that calling pop() will throw an exception now.
        Recommended to use before exiting the app or returning from an excetpion.
        """
        TealPrint._buffer.clear_indent()



import sys
import traceback
from io import StringIO
from threading import Lock
from typing import List

from colored import attr, fg

from . import TealConfig, TealLevel


class TealPrintBuffer:
    _mutex = Lock()
    _ascii: bool = False

    def __init__(self) -> None:
        self.buffer = StringIO()
        self.indent_stack: List[TealLevel] = []

    def error(
        self,
        message: str,
        push_indent: bool = False,
        exit: bool = False,
        print_exception: bool = False,
        print_report_this: bool = False,
    ) -> None:
        """Recommendation: Only use this when an error occurs and you have to exit the program.
           Add an error message in red to the buffer. Can print the exception.
           If exit=True, it flushes all messages before exiting.
           Optionally prints a "Please report this and paste the above message"
           Call flush() to print the messages.

        Args:
            message (str): The message to print
            push_indent (bool): If messages after this should be indented by one level. Call pop_indent() to unindent
            exit (bool): If the program should exit after printing the error. Also flushes messages
            print_exception (bool): Set to true to print an exception
            print_report_this (bool): Set to true to add an "Please report this..." message at the end
        """
        self._add_to_buffer_on_level(message, push_indent, fg("red"), TealLevel.error)
        if print_exception:
            exception = traceback.format_exc()
            self._add_to_buffer_on_level(exception, False, fg("red"), TealLevel.error)
        if print_report_this:
            self._add_to_buffer_on_level(
                "!!! Please report this and paste the above message !!!", False, fg("red"), TealLevel.error
            )
        if exit:
            self.flush()
            sys.exit(1)

    def warning(self, message: str, push_indent: bool = False, exit: bool = False) -> None:
        """Add an orange warning message to the buffer.
           If exit=True, it flushes all messages before exiting.
           Call flush() to print the messages.

        Args:
            message (str): The message to print
            push_indent (bool): If messages after this should be indented by one level. Call pop_indent() to unindent
            exit (bool): If the program should exit after printing the warning. Also flushes messages
        """
        self._add_to_buffer_on_level(message, push_indent, fg("dark_orange"), TealLevel.warning, exit)

    def info(self, message: str, push_indent: bool = False, color: str = "") -> None:
        """Add a message to the buffer if TealConfig.level has been set to debug/verbose/info.
           Call flush() to print the messages.

        Args:
            message (str): The message to print
            push_indent (bool): If messages after this should be indented by one level. Call pop_indent() to unindent
            color (str): Optional color of the message
        """
        self._add_to_buffer_on_level(message, push_indent, color, TealLevel.info)

    def verbose(self, message: str, push_indent: bool = False, color: str = "") -> None:
        """Add a message to the buffer if TealConfig.level has been set to debug/verbose.
           Call flush() to print the messages.

        Args:
            message (str): The message to print
            push_indent (bool): If messages after this should be indented by one level. Call pop_indent() to unindent
            color (str): Optional color of the message
        """
        self._add_to_buffer_on_level(message, push_indent, color, TealLevel.verbose)

    def debug(self, message: str, push_indent: bool = False, color: str = "") -> None:
        """Add a message to the buffer if the TealConfig.level has been set to debug.
           Call flush() to print the messages.

        Args:
            message (str): The message to print
            push_indent (bool): If messages after this should be indented by one level. Call pop_indent() to unindent
            color (str): Optional color of the message
        """
        self._add_to_buffer_on_level(message, push_indent, color, TealLevel.debug)

    def push_indent(self, level: TealLevel) -> None:
        """Add indentation for all messages, but only if level is would be shown"""
        self.indent_stack.append(level)

    def pop_indent(self) -> None:
        """
        Remove indentation for all messages. Note that nothing will happen if push_indent() was
        called with a level that wouldn't be shown.

        For example:
            TealConfig.level = TealLevel.info
            TealPrint.push_indent(TealLevel.warning)
            TealPrint.info("Will be indented by 1 level")
            TealPrint.push_indent(TealLevel.verbose)
            TealPrint.info("Will still be indented by 1 level")
            TealPrint.pop_indent()
            TealPrint.info("Will still be indented by 1 level...")
            TealPrint.pop_indent()
            TealPrint.info("Not indentend")
        """
        self.indent_stack.pop()

    def clear_indent(self) -> None:
        """
        Removes all indentation. Note that calling pop() will throw an exception now.
        Recommended to use before exiting the app or returning from an excetpion.
        """
        self.indent_stack.clear()

    def _add_to_buffer_on_level(
        self,
        message: str,
        push_indent: bool,
        color: str,
        level: TealLevel,
        exit: bool = False,
    ) -> None:
        """Prints the message if the level is equal or lower to the specified"""
        if TealConfig.level.value >= level.value:
            try:
                # Indent message
                indent_level = self._get_indent_level()
                if indent_level > 0:
                    message = "".ljust(indent_level * TealConfig.indent_by, TealConfig.indent_char) + message
                if len(color) > 0:
                    message = f"{color}{message}{attr('reset')}"

                # Convert to ascii if necessary
                if TealPrintBuffer._ascii:
                    message = message.encode("utf-8", "ignore").decode("ascii", "ignore")

                self._add_to_buffer(message)

                if exit:
                    self.flush()
                    sys.exit(1)

            except UnicodeEncodeError:
                # Some consoles can't use utf-8, encode into ascii instead, and use that
                # in the future
                TealPrintBuffer._ascii = True
                self._add_to_buffer_on_level(message, False, color, level, exit)

        # Always push indent
        if push_indent:
            self.push_indent(level)

    def _get_indent_level(self) -> int:
        count = 0
        for level in self.indent_stack:
            if TealConfig.level.value >= level.value:
                count += 1
        return count

    def _add_to_buffer(self, message: str) -> None:
        """Mostly used for mocking purposes"""
        self.buffer.write(message + "\n")

    def flush(self) -> None:
        """Prints the messages in the buffer"""
        self.buffer.seek(0)

        # Make sure we only print one message at a time
        TealPrintBuffer._mutex.acquire()
        try:
            print(self.buffer.read(), flush=True, end="")
        finally:
            TealPrintBuffer._mutex.release()

        self.buffer = StringIO()
